@@ -15,6 +15,9 @@
  * You should have received a copy of the GNU Lesser General
  * Public License along with this library; if not, see
  * <http://www.gnu.org/licenses/>.
+ *
+ * In addition, when the library is used with OpenSSL, a special
+ * exception applies. Refer to the LICENSE_EXCEPTION file for details.
  */
 
 #include "config.h"
@@ -375,6 +378,7 @@ g_tls_certificate_gnutls_real_copy (GTlsCertificateGnutls    *gnutls,
   gnutls_datum_t data;
   guint num_certs = 0;
   size_t size = 0;
+  int status;
 
   /* We will do this loop twice. It's probably more efficient than
    * re-allocating memory.
@@ -401,7 +405,8 @@ g_tls_certificate_gnutls_real_copy (GTlsCertificateGnutls    *gnutls,
                               data.data, &size);
 
       gnutls_x509_crt_init (&cert);
-      gnutls_x509_crt_import (cert, &data, GNUTLS_X509_FMT_DER);
+      status = gnutls_x509_crt_import (cert, &data, GNUTLS_X509_FMT_DER);
+      g_warn_if_fail (status == 0);
       g_free (data.data);
 
       st->cert.x509[st->ncerts] = cert;
